from llama.program.value import Value

from llama.program.util.type_to_dict import value_to_dict, type_to_dict

import json


class GetFieldOperation(Value):
    def __init__(self, input_value: Value, output_type: type, field_name: str):
        super().__init__(output_type)
        self._field_name = field_name
        self._input_value = input_value

    def _to_dict(self):
        input_value = value_to_dict(self._input_value)
        return {
            "name": "GetFieldOperation",
            "input_value": input_value,
            "type": type_to_dict(self._type),
            "field_name": self._field_name,
        }



from llama.engine.base_selector import BaseSelector


class ScoreSelector(BaseSelector):
    def __init__(self):
        super().__init__()
        self.higher_is_better = True
    
    def sort(self, examples, _):
        scored_examples, unscored_examples = [], []
        for e in examples:
            if 'score' in e:
                scored_examples.append(e)
            else:
                unscored_examples.append(e)
        sorted_examples = sorted(scored_examples, key=lambda elt: elt['score'], reverse=self.higher_is_better)
        examples = sorted_examples + unscored_examples
        return examples




from llama.program.value import Value

from llama.program.util.type_to_dict import value_to_dict

import json

class MetricOperation(Value):
    def __init__(self, input_value, output_type):
        super().__init__(output_type)
        self._input_value = input_value

    def to_dict(self):
        input_value = value_to_dict(self._input_value)

        return {
            "name" : "MetricOperation",
            "input_value" : input_value,
            "type" : json.loads(self._type.schema_json()),
        }





from llama.program.value import Value

import json


class CallOperation(Value):
    def __init__(self, target_function, input_value, output_value):
        super().__init__(get_type(output_value))
        self._target_function = target_function
        self._input_value = input_value

    def _to_dict(self):
        if isinstance(self._input_value, Value):
            input_value = self._input_value.index
        else:
            input_value = {
                "data": self._input_value.dict(),
                "type": json.loads(type(self._input_value).schema_json()),
            }

        return {
            "name": "CallOperation",
            "function_name": self._target_function.name,
            "input_value": input_value,
        }


def get_type(output_value):
    if type(output_value) == tuple:
        return (value.type for value in output_value)

    return output_value._type



from pydantic import BaseModel, validator
from pydantic.fields import ModelField

SHAPE_SINGLETON = 1
SHAPE_LIST = 2
SHAPE_SET = 3
SHAPE_MAPPING = 4
SHAPE_TUPLE = 5
SHAPE_TUPLE_ELLIPSIS = 6
SHAPE_SEQUENCE = 7
SHAPE_FROZENSET = 8
SHAPE_ITERABLE = 9
SHAPE_GENERIC = 10
SHAPE_DEQUE = 11
SHAPE_DICT = 12
SHAPE_DEFAULTDICT = 13
SHAPE_COUNTER = 14

class BaseSpecification(BaseModel):
    @validator('*', pre=True, always=True)
    def has_description(cls, v, field: ModelField):
        if field.field_info.description is None:
            if field.shape in (SHAPE_LIST, SHAPE_SET, SHAPE_MAPPING, SHAPE_TUPLE, SHAPE_DICT):
                raise ValueError(
                    f"Description missing for field {field.name} in Spec {cls.__name__}. Import Context from llama and use it to describe the field with natural language."
                )
            if field.type_ in (str, int, bool, float):
                raise ValueError(
                    f"Description missing for field {field.name} in Spec {cls.__name__}. Import Context from llama and use it to describe the field with natural language."
                )
        return v

    def _get_attribute_raw(self, name):
        return getattr(self, name)




from llama.program.value import Value
from llama.types.type import Type
from llama.types.context import Context

from llama.program.util.type_to_dict import value_to_dict

import json

class FeedbackType(Type):
    result: bool = Context("The result of the feedback test")

class FeedbackOperation(Value):
    def __init__(self, on, to, good_examples=[], bad_examples=[]):
        super().__init__(FeedbackType)
        self._on = on
        self._to = to
        self._good_examples = good_examples
        self._bad_examples = bad_examples

    def _to_dict(self):
        return {
            "name" : "FeedbackOperation",
            "on" : self._on,
            "to" : self._to,
            "good_examples" : [value_to_dict(example) for example in self._good_examples],
            "bad_examples" : [value_to_dict(example) for example in self._bad_examples],
            "type" : json.loads(self._type.schema_json()),
        }






import asyncio
from typing import List
from llama.program.util.run_ai import query_run_program

from llama.types.base_specification import BaseSpecification

class Value(object):
    def __init__(self, type, data=None):
        self._type = type
        self._data = data
        self._function = None
        self._index = None

    def _get_field(self, name):
        if self._data is None:
            self._compute_value()

        return self._data._get_attribute_raw(name)

    def __str__(self):
        if self._data is None:
            self._compute_value()

        return str(self._data)

    def __int__(self):
        if self._data is None:
            self._compute_value()

        return int(self._data)

    def __float__(self):
        if self._data is None:
            self._compute_value()

        return float(self._data)

    def __gt__(self, other):
        if self._data is None:
            self._compute_value()

        if isinstance(other, Value):
            other = other._get_data()

        return self._data > other

    def _get_data(self):
        if self._data is None:
            self._compute_value()

        return self._data

    def _compute_value(self):
        # check in the builper value cache
        if self._index in self._function.program.builder.value_cache:
            returned_value = self._function.program.builder.value_cache[self._index]["data"]
        else:
            params = {
                "program": self._function.program.to_dict(),
                "requested_values": [self._index],
            }

            response = query_run_program(params)

            response.raise_for_status()

            # update the cache
            self._function.program.builder.value_cache.update(response.json())

            returned_value = response.json()[str(self._index)]["data"]

        if issubclass(self._type, BaseSpecification):
            self._data = self._type.parse_obj(returned_value)
        else:
            self._data = self._type(returned_value)

    def __getattribute__(self, name):
        if name.find("_") == 0:
            return super().__getattribute__(name)

        return self._function.program.builder.get_field(self, name)

    def _get_attribute_raw(self, name):
        return super().__getattribute__(name)



def gen_multiple_values(values: List[Value]):
    program = values[0].function.program.to_dict()
    params = {
        "program": program,
        "requested_value": [v.index for v in values],
    }

    response = query_run_program(params)
    response.raise_for_status()
    for i, v in enumerate(values):
        v.data = v.type.parse_obj(response.json()[i]["data"])
    return values







from llama.program.program import Program
from llama.program.function import Function
from llama.program.util.run_ai import query_run_program
from llama.program.value import Value

from llama.types.type import Type

from llama.program.operations.llama_operation import LlamaOperation
from llama.program.operations.metric_operation import MetricOperation
from llama.program.operations.call_operation import CallOperation
from llama.program.operations.get_element_operation import GetElementOperation
from llama.program.operations.get_field_operation import GetFieldOperation
from llama.program.operations.return_operation import ReturnOperation
from llama.program.operations.feedback_operation import FeedbackOperation

from llama.types.base_specification import BaseSpecification

import requests

import inspect


class Builder:
    """Build a program for execution by the Llama large language model engine."""

    def __init__(self, name):
        self.program = Program(self, name)
        self.current_function = self.program.main
        self.value_cache = {}

    def __call__(self, input, output_type, *args, **kwargs):
        input_value = input
        new_operation = self.current_function.add_operation(
            LlamaOperation(input_value, output_type, *args, **kwargs)
        )

        return new_operation

    def fit(self, examples=[]):
        self.add_data(examples)

    def add_data(self, data=[]):
        self.program.add_data(examples=data)

    def improve(self, on: str, to: str, good_examples=[], bad_examples=[]):

        new_operation = self.current_function.add_operation(
            FeedbackOperation(
                on=on, to=to, good_examples=good_examples, bad_examples=bad_examples
            )
        )

        return new_operation

    def function(self, function):
        signature = inspect.signature(function)
        input_types = [value.annotation for value in signature.parameters.values()]

        main = self.current_function
        new_function = Function(
            program=self.program, name=function.__name__, input_arguments=input_types
        )
        self.program.functions[new_function.name] = new_function
        self.current_function = new_function
        output_value = function(*new_function.operations)
        self.current_function.add_operation(ReturnOperation(output_value))
        self.current_function = main

        return Lambda(self, new_function, output_value)

    def add_call(self, function, input_value, output_value):
        new_operation = self.current_function.add_operation(
            CallOperation(function, input_value, output_value)
        )

        result = new_operation

        if type(output_value) is tuple:
            result = []

            for index, value in enumerate(output_value):
                result.append(
                    self.current_function.add_operation(
                        GetElementOperation(new_operation, value.type, index)
                    )
                )

        return result

    def get_field(self, value, field_name):
        return self.current_function.add_operation(
            GetFieldOperation(value, value._type._get_field_type(field_name), field_name)
        )

    def add_metric(self, metric):
        new_operation = self.current_function.add_operation(
            MetricOperation(metric.input, metric.get_metric_type())
        )

        return new_operation

    def make_metric(
        self, input: Type, metric_type: type, fit: bool = True, higher_is_better=True
    ):
        new_operation = self.current_function.add_operation(
            MetricOperation(input, metric_type)
        )

        return new_operation

    def metrics(self):
        requested_values = [
            op._index for op in self.program.functions["main"].operations
        ]

        params = {
            "program": self.program.to_dict(),
            "requested_values": requested_values,
        }
        response = query_run_program(params)
        response.raise_for_status()

        data = [result[str(index)]["data"] for index in requested_values]

        return data


class Lambda:
    def __init__(self, builder: Builder, function: Function, output_value: Type):
        self.output_value = output_value
        self.builder = builder
        self.function = function

    def __call__(self, *args, **kwargs):
        input_value = self._get_input(*args, **kwargs)
        return self.builder.add_call(self.function, input_value, self.output_value)

    def _get_input(self, *args, **kwargs):
        # TODO: support more than one input LLM arg

        if len(args) > 0:
            return args[0]

        return next(iter(kwargs.values()))




from llama.program.value import Value

from llama.program.util.type_to_dict import value_to_dict, type_to_dict

import json

class GetElementOperation(Value):
    def __init__(self, input_value, type, element_index):
        super().__init__(type)
        self._element_index = element_index
        self._input_value = input_value

    def _to_dict(self):
        input_value = type_to_dict(self._input_value)
        return {
            "name" : "GetElementOperation",
            "input_value" : input_value,
            "type" : type_to_dict(self._type),
            "element_index" : self._element_index,
        }




from llama.program.value import Value

class GetArgumentOperation(Value):
    def __init__(self, type, input_value_index):
        super().__init__(type)
        self._input_value_index = input_value_index

    def _to_dict(self):
        return {
            "name" : "GetArgumentOperation",
            "index" : self._input_value_index
        }




import openai

# TODO: confirm what cue works best for lists
# LIST_CUE = '['
# LIST_CUE = '1.'
LIST_CUE = "\n-"
ESC_LIST_CUE = "\\n-"
SEPARATOR = "\n"
CHUNK_SEPARATOR = " (continuing input) "
EXAMPLE_SEPARATOR = "END"
STOP_TOKEN = "\nEND"
FIXED_TEMPERATURE = 0.0
RANDOM_TEMPERATURE = 0.7
MAX_INPUT_TOKENS = 3872
MAX_OUTPUT_TOKENS = 256
EMBEDDING_MODEL = lambda prompt: openai.Embedding.create(
    input=[prompt], model="text-embedding-ada-002"
)["data"][0]["embedding"]



import requests
import os
from llama.program.util.config import get_config


def query_run_program(params):
    key, url = get_url_and_key()
    resp = powerml_run_program(params, url, key)
    return resp

def powerml_run_program(params, url, key):
    headers = {
        'Content-Type': 'application/json',
        'Authorization': 'Bearer ' + key,
    }
    response = requests.post(
        url=url + "/v1/run_llama_program",
        headers=headers,
        json=params)
    if response.status_code != 200:
        try:
            description = response.json()
            print(description)
        except BaseException:
            description = response.status_code
        finally:
            raise Exception(f"API error {description}")
    return response

def get_url_and_key():
    cfg = get_config()
    environment = os.environ.get("LLAMA_ENVIRONMENT")
    if environment == "LOCAL":
        key = 'test_token'
        if 'local' in cfg:
            if 'key' in cfg["local"]:
                key = cfg['local.key']
        url = "http://localhost:5001"
    elif environment == "STAGING":
        key = cfg['staging.key']
        url = 'https://api.staging.powerml.co'
    else:
        key = cfg['production.key']
        url = 'https://api.powerml.co'
    return (key, url)




from llama.program.operations.get_argument_operation import GetArgumentOperation

class Function:
    def __init__(self, program, name, input_arguments=[]):
        self.name = name
        self.program=program
        self.operations = []

        for index, argument in enumerate(input_arguments):
            self.add_operation(GetArgumentOperation(argument, index))

    def add_operation(self, operation):
        operation._index = len(self.operations)
        operation._function = self

        self.operations.append(operation)

        return operation

    def to_dict(self):
        dict_object = {
            "name" : self.name,
            "operations" : [operation._to_dict() for operation in self.operations]
        }
        return dict_object





from llama import Type, Context, LLM

import unittest

class Story(Type):
    story: str = Context("the body of the story")


class Tone(Type):
    tone: str = Context("The tone of the story")


class Descriptors(Type):
    likes: str = Context("things you like")
    favorite_song: str = Context("your favorite song")
    tone: Tone = Context("tone of the story")


llm = LLM(name="test_random")

@llm.function
def write_story(descriptors: Descriptors) -> Story:
    story = llm(
        input=descriptors, output_type=Story, random=True
    )

    return story

class TestRandom(unittest.TestCase):
    def test_random(self):
        descriptors = Descriptors(
            likes="llamas and other animals",
            favorite_song="never let me go",
            tone=Tone(tone="cheeky"),
        )

        story = write_story(descriptors)

        print(story)






from llama.types.base_specification import BaseSpecification
from llama.program.value import Value

from pydantic import PrivateAttr
import inspect
import typing


class Type(BaseSpecification):
    _value = PrivateAttr()

    def __init__(self, *args, **kwargs):
        if any_values(args, kwargs):
            unvalidated = super().construct(*args, **kwargs)
            object.__setattr__(self, '__dict__', unvalidated.__dict__)
            object.__setattr__(self, '__fields_set__', unvalidated.__fields_set__)
        else:
            super().__init__(*args, **kwargs)

        self._value = Value(type(self), data=self)

    def __getattribute__(self, name):
        if name.find("_") == 0:
            if name != "_value":
                return super().__getattribute__(name)

        # fix access to private fields
        members = inspect.getmembers(Type)
        for member_name, member in members:
            if member_name == "__dict__":
                value = member["_value"].__get__(self)

        if name == "_value":
            return value

        if name in self.__dict__:
            return value._get_field(name)

        return super().__getattribute__(name)

    def _get_attribute_raw(self, name):
        return super().__getattribute__(name)

    @classmethod
    def _get_field_type(cls, name):
        field_to_type = typing.get_type_hints(cls)
        return field_to_type[name]

def any_values(args, kwargs):
    for arg in args:
        if isinstance(arg, Value):
            return True

    for arg in kwargs.values():
        if isinstance(arg, Value):
            return True

    return False





from llama.program.value import Value

class ReturnOperation(Value):
    def __init__(self, output_value):
        super().__init__(get_type(output_value))
        self._output_value = output_value

    def _to_dict(self):
        if type(self._output_value) == tuple:
            output_value = [value_to_dict(value) for value in self._output_value]
        else:
            output_value = value_to_dict(self._output_value)

        return {
            "name" : "ReturnOperation",
            "output_value" : output_value,
        }

def value_to_dict(value):
    if isinstance(value, Value):
        value = value._index
    else:
        value = value.dict()

    return value

def get_type(output_value):
    if type(output_value) == tuple:
        return (value._type for value in output_value)

    return output_value._type





from llama.program.value import Value
from llama.types.base_specification import BaseSpecification

import json


def type_to_dict(type):
    if type == BaseSpecification:
        return json.loads(self._type.schema_json()),

    return { "type" : str(type) }

def value_to_dict(input_value):
    if isinstance(input_value, Value):
        return input_value._index

    return {
        "data": rewrite_dict_data(input_value),
        "type": rewrite_dict_type(input_value),
    }


def rewrite_dict_data(input_value):
    if isinstance(input_value, Value):
        return input_value._index
    elif isinstance(input_value, BaseSpecification):
        input_value = input_value.dict()
        for key, value in input_value.items():
            input_value[key] = rewrite_dict_data(value)

    return input_value


def rewrite_dict_type(input_value):
    assert isinstance(input_value, BaseSpecification)
    return json.loads(type(input_value).schema_json())




from llama.metrics.metric import Metric
from llama.types.type import Type
from llama.types.context import Context

class Match(Type):
    a: str = Context("the first string")
    b: str = Context("the second string")

class MatchResult(Type):
    similarity: float = Context(
        "a number between 0.0 and 1.0 describing similarity of the input strings.  0.0 is no similarity, and 1.0 is a perfect match."
    )

class CompareEqualMetric(Metric):
    def __init__(self, a, b):
        self.input = Match(a=a, b=b)

        print("match", self.input)

    def get_metric_type(self):
        return MatchResult




from llama.program.function import Function
from llama.program.util.type_to_dict import value_to_dict


class Program:
    """Internal representation of a program that can be executed by the Llama
    large language model engine.

    Each program has a unique name (within your account).

    """

    def __init__(self, builder, name):
        self.builder = builder
        self.name = name
        self.main = Function(program=self, name="main")
        self.functions = {"main": self.main}
        self.examples = []

    def add_data(self, examples):
        self.examples.extend(examples)

    def add_metric(self, metric):
        self.add_operation(metric)

    def to_dict(self):
        dict_object = {
            "name": self.name,
            "functions": {
                name: function.to_dict() for name, function in self.functions.items()
            },
            "examples": [
                {
                    "input": value_to_dict(example["input"]),
                    "output": value_to_dict(example["output"]),
                }
                for example in self.examples
            ],
        }

        return dict_object
